"""Constants for the PyISY Module."""
import datetime

UPDATE_INTERVAL = 0.5

# Time Constants / Strings
EMPTY_TIME = datetime.datetime(year=1, month=1, day=1)
ISY_EPOCH_OFFSET = 36524
MILITARY_TIME = "%Y/%m/%d %H:%M:%S"
STANDARD_TIME = "%Y/%m/%d %I:%M:%S %p"
XML_STRPTIME = "%Y%m%d %H:%M:%S"
XML_STRPTIME_YY = "%y%m%d %H:%M:%S"
XML_TRUE = "true"
XML_FALSE = "false"
XML_ON = "<on />"
XML_OFF = "<off />"

POLL_TIME = 10
RECONNECT_DELAY = 60
SOCKET_BUFFER_SIZE = 4096
THREAD_SLEEP_TIME = 30.0

ES_LOST_STREAM_CONNECTION = "lost_stream_connection"
ES_CONNECTED = "connected"
ES_DISCONNECTED = "disconnected"
ES_START_UPDATES = "start_updates"
ES_STOP_UPDATES = "stop_updates"
ES_INITIALIZING = "stream_initializing"
ES_LOADED = "stream_loaded"
ES_RECONNECT_FAILED = "reconnect_failed"
ES_RECONNECTING = "reconnecting"
ES_DISCONNECTING = "stream_disconnecting"
ES_NOT_STARTED = "not_started"

ISY_VALUE_UNKNOWN = -1 * float("inf")
ISY_PROP_NOT_SET = "-1"

""" Dictionary of X10 commands. """
X10_COMMANDS = {"all_off": 1, "all_on": 4, "on": 3, "off": 11, "bright": 7, "dim": 15}

ACTION_EVENT_STATUS = "0"
ACTION_GET_STATUS = "1"
ACTION_KEY_CHANGED = "2"
ACTION_INFO_STRING = "3"
ACTION_IR_LEARN = "4"
ACTION_SCHEDULE = "5"
ACTION_VAR_STATUS = "6"
ACTION_VAR_INIT = "7"
ACTION_KEY = "8"

ATTR_ACTION = "action"
ATTR_CONTROL = "control"
ATTR_DESC = "desc"
ATTR_FLAG = "flag"
ATTR_FORMATTED = "formatted"
ATTR_ID = "id"
ATTR_INIT = "init"
ATTR_INSTANCE = "instance"
ATTR_LAST_CHANGED = "last_changed"
ATTR_LAST_UPDATE = "last_update"
ATTR_NODE_DEF_ID = "nodeDefId"
ATTR_PARENT = "parentId"
ATTR_PRECISION = "prec"
ATTR_SET = "set"
ATTR_STATUS = "status"
ATTR_STREAM_ID = "sid"
ATTR_TS = "ts"
ATTR_UNIT_OF_MEASURE = "uom"
ATTR_VAL = "val"  # Used for Variables.
ATTR_VALUE = "value"  # Used for everything else.
ATTR_VAR = "var"

DEFAULT_PRECISION = "0"
DEFAULT_UNIT_OF_MEASURE = ""

TAG_ADDRESS = "address"
TAG_CATEGORY = "cat"
TAG_CONFIG = "config"
TAG_DESC = "desc"
TAG_DESCRIPTION = "description"
TAG_DEVICE_TYPE = "devtype"
TAG_DST = "DST"
TAG_ENABLED = "enabled"
TAG_EVENT_INFO = "eventInfo"
TAG_FAMILY = "family"
TAG_FEATURE = "feature"
TAG_FIRMWARE = "app_full_version"
TAG_FOLDER = "folder"
TAG_FORMATTED = "fmtAct"
TAG_GENERIC = "gen"
TAG_GROUP = "group"
TAG_INSTALLED = "isInstalled"
TAG_IS_LOAD = "isLoad"
TAG_LATITUDE = "Lat"
TAG_LINK = "link"
TAG_LOCATION = "location"
TAG_LONGITUDE = "Long"
TAG_MFG = "mfg"
TAG_MILIATRY_TIME = "IsMilitary"
TAG_NAME = "name"
TAG_NET_RULE = "NetRule"
TAG_NODE = "node"
TAG_NODE_DEFS = "nodedefs"
TAG_NTP = "NTP"
TAG_PARENT = "parent"
TAG_PARAMETER = "parameter"
TAG_PRGM_FINISH = "f"
TAG_PRGM_RUN = "r"
TAG_PRGM_RUNNING = "running"
TAG_PRGM_STATUS = "s"
TAG_PRIMARY_NODE = "pnode"
TAG_PRODUCT = "product"
TAG_PROGRAM = "program"
TAG_PROPERTY = "property"
TAG_ROOT = "root"
TAG_SIZE = "size"
TAG_SPOKEN = "spoken"
TAG_SUNRISE = "Sunrise"
TAG_SUNSET = "Sunset"
TAG_TYPE = "type"
TAG_TZ_OFFSET = "TMZOffset"
TAG_VALUE = "value"
TAG_VARIABLE = "e"
TAG_VARIABLES = "variables"

PROTO_FOLDER = "program_folder"
PROTO_GROUP = "group"
PROTO_INSTEON = "insteon"
PROTO_INT_VAR = "integer_variable"
PROTO_ISY = "isy"
PROTO_NETWORK = "network"
PROTO_NODE_SERVER = "node_server"
PROTO_PROGRAM = "program"
PROTO_STATE_VAR = "state_variable"
PROTO_ZIGBEE = "zigbee"
PROTO_ZWAVE = "zwave"

FAMILY_CORE = "0"
FAMILY_INSTEON = "1"
FAMILY_UPB = "2"
FAMILY_RCS = "3"
FAMILY_ZWAVE = "4"
FAMILY_AUTO = "5"
FAMILY_GENERIC = "6"
FAMILY_UDI = "7"
FAMILY_BRULTECH = "8"
FAMILY_NCD = "9"
FAMILY_NODESERVER = "10"
FAMILY_ZMATTER_ZWAVE = "12"

PROP_BATTERY_LEVEL = "BATLVL"
PROP_BUSY = "BUSY"
PROP_COMMS_ERROR = "ERR"
PROP_ENERGY_MODE = "CLIEMD"
PROP_HEAT_COOL_STATE = "CLIHCS"
PROP_HUMIDITY = "CLIHUM"
PROP_ON_LEVEL = "OL"
PROP_RAMP_RATE = "RR"
PROP_SCHEDULE_MODE = "CLISMD"
PROP_SETPOINT_COOL = "CLISPC"
PROP_SETPOINT_HEAT = "CLISPH"
PROP_STATUS = "ST"
PROP_TEMPERATURE = "CLITEMP"
PROP_UOM = "UOM"
PROP_ZWAVE_PREFIX = "ZW_"

METHOD_COMMAND = "cmd"
METHOD_GET = "get"
METHOD_SET = "set"

URL_CHANGE = "change"
URL_CLOCK = "time"
URL_CONFIG = "config"
URL_DEFINITIONS = "definitions"
URL_MEMBERS = "members"
URL_NETWORK = "networking"
URL_NODE = "node"
URL_NODES = "nodes"
URL_NOTES = "notes"
URL_PING = "ping"
URL_PROGRAMS = "programs"
URL_QUERY = "query"
URL_RESOURCES = "resources"
URL_STATUS = "status"
URL_SUBFOLDERS = "subfolders"
URL_VARIABLES = "vars"
URL_ZWAVE = "zwave"
URL_PROFILE_NS = "profiles/ns"

VAR_INTEGER = "1"
VAR_STATE = "2"

CLIMATE_SETPOINT_MIN_GAP = 2

CMD_BACKLIGHT = "BL"
CMD_BEEP = "BEEP"
CMD_BRIGHTEN = "BRT"
CMD_CLIMATE_FAN_SETTING = "CLIFS"
CMD_CLIMATE_MODE = "CLIMD"
CMD_DIM = "DIM"
CMD_DISABLE = "disable"
CMD_DISABLE_RUN_AT_STARTUP = "disableRunAtStartup"
CMD_ENABLE = "enable"
CMD_ENABLE_RUN_AT_STARTUP = "enableRunAtStartup"
CMD_FADE_DOWN = "FDDOWN"
CMD_FADE_STOP = "FDSTOP"
CMD_FADE_UP = "FDUP"
CMD_MANUAL_DIM_BEGIN = "BMAN"  # Depreciated, use Fade
CMD_MANUAL_DIM_STOP = "SMAN"  # Depreciated, use Fade
CMD_MODE = "MODE"
CMD_OFF = "DOF"
CMD_OFF_FAST = "DFOF"
CMD_ON = "DON"
CMD_ON_FAST = "DFON"
CMD_RESET = "RESET"
CMD_RUN = "run"
CMD_RUN_ELSE = "runElse"
CMD_RUN_THEN = "runThen"
CMD_SECURE = "SECMD"
CMD_STOP = "stop"
CMD_X10 = "X10"

COMMAND_FRIENDLY_NAME = {
    "ADRPST": "auto_dr_processing_state",
    "AIRFLOW": "air_flow",
    "ALARM": "alarm",
    "ANGLE": "angle_position",
    "ANGLPOS": "angle_position",
    "ATMPRES": "atmospheric_pressure",
    "AWAKE": "awake",
    "BARPRES": "barometric_pressure",
    "CC": "current",
    "CLIFRS": "fan_running_state",
    "CLIFSO": "fan_setting_override",
    "CO2LVL": "co2_level",
    "CPW": "power",
    "CTL": "controller_action",
    "CV": "voltage",
    "DELAY": "delay",
    "DEWPT": "dew_point",
    "DISTANC": "distance",
    "DOF3": "off_3x_key_presses",
    "DOF4": "off_4x_key_presses",
    "DOF5": "off_5x_key_presses",
    "DON3": "on_3x_key_presses",
    "DON4": "on_4x_key_presses",
    "DON5": "on_5x_key_presses",
    "ELECCON": "electrical_conductivity",
    "ELECRES": "electrical_resistivity",
    PROP_COMMS_ERROR: "device_communication_errors",
    "ETO": "evapotranspiration",
    "FATM": "fat_mass",
    "FREQ": "frequency",
    "GPV": "general_purpose",
    "GUST": "gust",
    "GV0": "custom_control_0",
    "GV1": "custom_control_1",
    "GV2": "custom_control_2",
    "GV3": "custom_control_3",
    "GV4": "custom_control_4",
    "GV5": "custom_control_5",
    "GV6": "custom_control_6",
    "GV7": "custom_control_7",
    "GV8": "custom_control_8",
    "GV9": "custom_control_9",
    "GV10": "custom_control_10",
    "GV11": "custom_control_11",
    "GV12": "custom_control_12",
    "GV13": "custom_control_13",
    "GV14": "custom_control_14",
    "GV15": "custom_control_15",
    "GV16": "custom_control_16",
    "GV17": "custom_control_17",
    "GV18": "custom_control_18",
    "GV19": "custom_control_19",
    "GV20": "custom_control_20",
    "GV21": "custom_control_21",
    "GV22": "custom_control_22",
    "GV23": "custom_control_23",
    "GV24": "custom_control_24",
    "GV25": "custom_control_25",
    "GV26": "custom_control_26",
    "GV27": "custom_control_27",
    "GV28": "custom_control_28",
    "GV29": "custom_control_29",
    "GV30": "custom_control_30",
    "GVOL": "gas_volume",
    "HAIL": "hail",
    "HEATIX": "heat_index",
    "HR": "heart_rate",
    "LUMIN": "luminance",
    "METHANE": "methane_density",
    "MOIST": "moisture",
    "MOON": "moon_phase",
    "MUSCLEM": "muscle_mass",
    "OZONE": "ozone",
    "PCNT": "pulse_count",
    "PF": "power_factor",
    "PM10": "particulate_matter_10",
    "PM25": "particulate_matter_2.5",
    "POP": "percent_chance_of_precipitation",
    "PPW": "polarized_power",
    "PRECIP": "precipitation",
    "PULSCNT": "pulse_count",
    "RADON": "radon_concentration",
    "RAINRT": "rain_rate",
    "RELMOD": "relative_modulation_level",
    "RESPR": "respiratory_rate",
    "RFSS": "rf_signal_strength",
    "ROTATE": "rotation",
    "RR": "ramp_rate",
    "SEISINT": "seismic_intensity",
    "SEISMAG": "seismic_magnitude",
    "SMOKED": "smoke_density",
    "SOILH": "soil_humidity",
    "SOILR": "soil_reactivity",
    "SOILS": "soil_salinity",
    "SOILT": "soil_temperature",
    "SOLRAD": "solar_radiation",
    "SPEED": "speed",
    "SVOL": "sound_volume",
    "TANKCAP": "tank_capacity",
    "TEMPEXH": "exhaust_temperature",
    "TEMPOUT": "outside_temperature",
    "TIDELVL": "tide_level",
    "TIME": "time",
    "TIMEREM": "time_remaining",
    "TPW": "total_energy_used",
    "UAC": "user_number",
    "USRNUM": "user_number",
    "UV": "uv_light",
    "VOCLVL": "voc_level",
    "WATERF": "water_flow",
    "WATERP": "water_pressure",
    "WATERT": "water_temperature",
    "WATERTB": "boiler_water_temperature",
    "WATERTD": "domestic_hot_water_temperature",
    "WEIGHT": "weight",
    "WINDCH": "wind_chill",
    "WINDDIR": "wind_direction",
    "WVOL": "water_volume",
    CMD_BACKLIGHT: "backlight",
    CMD_BEEP: "beep",
    CMD_BRIGHTEN: "bright",
    CMD_CLIMATE_FAN_SETTING: "fan_state",
    CMD_CLIMATE_MODE: "climate_mode",
    CMD_DIM: "dim",
    CMD_FADE_DOWN: "fade_down",
    CMD_FADE_STOP: "fade_stop",
    CMD_FADE_UP: "fade_up",
    CMD_MANUAL_DIM_BEGIN: "brighten_manual",
    CMD_MANUAL_DIM_STOP: "stop_manual",
    CMD_MODE: "mode",
    CMD_OFF: "off",
    CMD_OFF_FAST: "fastoff",
    CMD_ON: "on",
    CMD_ON_FAST: "faston",
    CMD_RESET: "reset",
    CMD_SECURE: "secure",
    CMD_X10: "x10_command",
    PROP_BATTERY_LEVEL: "battery_level",
    PROP_BUSY: "busy",
    PROP_ENERGY_MODE: "energy_saving_mode",
    PROP_HEAT_COOL_STATE: "heat_cool_state",
    PROP_HUMIDITY: "humidity",
    PROP_ON_LEVEL: "on_level",
    PROP_SCHEDULE_MODE: "schedule_mode",
    PROP_SETPOINT_COOL: "cool_setpoint",
    PROP_SETPOINT_HEAT: "heat_setpoint",
    PROP_STATUS: "status",
    PROP_TEMPERATURE: "temperature",
    PROP_UOM: "unit_of_measure",
}

EVENT_PROPS_IGNORED = [
    CMD_BEEP,
    CMD_BRIGHTEN,
    CMD_DIM,
    CMD_MANUAL_DIM_BEGIN,
    CMD_MANUAL_DIM_STOP,
    CMD_FADE_UP,
    CMD_FADE_DOWN,
    CMD_FADE_STOP,
    CMD_OFF,
    CMD_OFF_FAST,
    CMD_ON,
    CMD_ON_FAST,
    CMD_RESET,
    CMD_X10,
    PROP_BUSY,
    PROP_STATUS,
]

COMMAND_NAME = {val: key for key, val in COMMAND_FRIENDLY_NAME.items()}

# Referenced from ISY-WSDK-5.0.4\WSDL\family.xsd
NODE_FAMILY_ID = {
    FAMILY_CORE: "Default",
    FAMILY_INSTEON: "Insteon",
    FAMILY_UPB: "UPB",
    FAMILY_RCS: "RCS",
    FAMILY_ZWAVE: "Z-Wave",
    FAMILY_AUTO: "Auto_DR",
    FAMILY_GENERIC: "Group",
    FAMILY_UDI: "UDI",
    FAMILY_BRULTECH: "Brultech",
    FAMILY_NCD: "NCD",
    FAMILY_NODESERVER: "Node_Server",
    FAMILY_ZMATTER_ZWAVE: "ZMatter_Z-Wave",
}

UOM_CLIMATE_MODES = "98"
UOM_CLIMATE_MODES_ZWAVE = "67"
UOM_DOUBLE_TEMP = "101"
UOM_FAN_MODES = "99"
UOM_INDEX = "25"
UOM_PERCENTAGE = "51"
UOM_SECONDS = "57"

UOM_FRIENDLY_NAME = {
    "1": "A",
    "2": "",  # Binary / On-Off
    "3": "btu/h",
    "4": "°C",
    "5": "cm",
    "6": "ft³",
    "7": "ft³/min",
    "8": "m³",
    "9": "day",
    "10": "days",
    "12": "dB",
    "13": "dB A",
    "14": "°",
    "16": "macroseismic",
    "17": "°F",
    "18": "ft",
    "19": "hour",
    "20": "hours",
    "21": "%AH",
    "22": "%RH",
    "23": "inHg",
    "24": "in/hr",
    UOM_INDEX: "index",
    "26": "K",
    "27": "keyword",
    "28": "kg",
    "29": "kV",
    "30": "kW",
    "31": "kPa",
    "32": "KPH",
    "33": "kWh",
    "34": "liedu",
    "35": "L",
    "36": "lx",
    "37": "mercalli",
    "38": "m",
    "39": "m³/hr",
    "40": "m/s",
    "41": "mA",
    "42": "ms",
    "43": "mV",
    "44": "min",
    "45": "min",
    "46": "mm/hr",
    "47": "month",
    "48": "MPH",
    "49": "m/s",
    "50": "Ω",
    UOM_PERCENTAGE: "%",
    "52": "lbs",
    "53": "pf",
    "54": "ppm",
    "55": "pulse count",
    "57": "s",
    "58": "s",
    "59": "S/m",
    "60": "m_b",
    "61": "M_L",
    "62": "M_w",
    "63": "M_S",
    "64": "shindo",
    "65": "SML",
    "69": "gal",
    "71": "UV index",
    "72": "V",
    "73": "W",
    "74": "W/m²",
    "75": "weekday",
    "76": "°",
    "77": "year",
    "82": "mm",
    "83": "km",
    "85": "Ω",
    "86": "kΩ",
    "87": "m³/m³",
    "88": "Water activity",
    "89": "RPM",
    "90": "Hz",
    "91": "°",
    "92": "° South",
    "100": "",
    "101": "° (x2)",
    "102": "kWs",
    "103": "$",
    "104": "¢",
    "105": "in",
    "106": "mm/day",
    "107": "",  # raw 1-byte unsigned value
    "108": "",  # raw 2-byte unsigned value
    "109": "",  # raw 3-byte unsigned value
    "110": "",  # raw 4-byte unsigned value
    "111": "",  # raw 1-byte signed value
    "112": "",  # raw 2-byte signed value
    "113": "",  # raw 3-byte signed value
    "114": "",  # raw 4-byte signed value
    "116": "mi",
    "117": "mbar",
    "118": "hPa",
    "119": "Wh",
    "120": "in/day",
}

UOM_TO_STATES = {
    "11": {  # Deadbolt Status
        "0": "unlocked",
        "100": "locked",
        "101": "unknown",
        "102": "problem",
    },
    "15": {  # Door Lock Alarm
        "1": "master code changed",
        "2": "tamper code entry limit",
        "3": "escutcheon removed",
        "4": "key-manually locked",
        "5": "locked by touch",
        "6": "key-manually unlocked",
        "7": "remote locking jammed bolt",
        "8": "remotely locked",
        "9": "remotely unlocked",
        "10": "deadbolt jammed",
        "11": "battery too low to operate",
        "12": "critical low battery",
        "13": "low battery",
        "14": "automatically locked",
        "15": "automatic locking jammed bolt",
        "16": "remotely power cycled",
        "17": "lock handling complete",
        "19": "user deleted",
        "20": "user added",
        "21": "duplicate pin",
        "22": "jammed bolt by locking with keypad",
        "23": "locked by keypad",
        "24": "unlocked by keypad",
        "25": "keypad attempt outside schedule",
        "26": "hardware failure",
        "27": "factory reset",
        "28": "manually not fully locked",
        "29": "all user codes deleted",
        "30": "new user code not added-duplicate code",
        "31": "keypad temporarily disabled",
        "32": "keypad busy",
        "33": "new program code entered",
        "34": "rf unlock with invalid user code",
        "35": "rf lock with invalid user codes",
        "36": "window-door is open",
        "37": "window-door is closed",
        "38": "window-door handle is open",
        "39": "window-door handle is closed",
        "40": "user code entered on keypad",
        "41": "power cycled",
    },
    "66": {  # Thermostat Heat/Cool State
        "0": "idle",
        "1": "heating",
        "2": "cooling",
        "3": "fan_only",
        "4": "pending heat",
        "5": "pending cool",
        "6": "vent",
        "7": "aux heat",
        "8": "2nd stage heating",
        "9": "2nd stage cooling",
        "10": "2nd stage aux heat",
        "11": "3rd stage aux heat",
    },
    "67": {  # Thermostat Mode
        "0": "off",
        "1": "heat",
        "2": "cool",
        "3": "auto",
        "4": "aux/emergency heat",
        "5": "resume",
        "6": "fan_only",
        "7": "furnace",
        "8": "dry air",
        "9": "moist air",
        "10": "auto changeover",
        "11": "energy save heat",
        "12": "energy save cool",
        "13": "away",
        "14": "program auto",
        "15": "program heat",
        "16": "program cool",
    },
    "68": {  # Thermostat Fan Mode
        "0": "auto",
        "1": "on",
        "2": "auto high",
        "3": "high",
        "4": "auto medium",
        "5": "medium",
        "6": "circulation",
        "7": "humidity circulation",
        "8": "left-right circulation",
        "9": "up-down circulation",
        "10": "quiet",
    },
    "78": {"0": "off", "100": "on"},  # 0-Off 100-On
    "79": {"0": "open", "100": "closed"},  # 0-Open 100-Close
    "80": {  # Thermostat Fan Run State
        "0": "off",
        "1": "on",
        "2": "on high",
        "3": "on medium",
        "4": "circulation",
        "5": "humidity circulation",
        "6": "right/left circulation",
        "7": "up/down circulation",
        "8": "quiet circulation",
    },
    "84": {"0": "unlock", "1": "lock"},  # Secure Mode
    "93": {  # Power Management Alarm
        "1": "power applied",
        "2": "ac mains disconnected",
        "3": "ac mains reconnected",
        "4": "surge detection",
        "5": "volt drop or drift",
        "6": "over current detected",
        "7": "over voltage detected",
        "8": "over load detected",
        "9": "load error",
        "10": "replace battery soon",
        "11": "replace battery now",
        "12": "battery is charging",
        "13": "battery is fully charged",
        "14": "charge battery soon",
        "15": "charge battery now",
    },
    "94": {  # Appliance Alarm
        "1": "program started",
        "2": "program in progress",
        "3": "program completed",
        "4": "replace main filter",
        "5": "failure to set target temperature",
        "6": "supplying water",
        "7": "water supply failure",
        "8": "boiling",
        "9": "boiling failure",
        "10": "washing",
        "11": "washing failure",
        "12": "rinsing",
        "13": "rinsing failure",
        "14": "draining",
        "15": "draining failure",
        "16": "spinning",
        "17": "spinning failure",
        "18": "drying",
        "19": "drying failure",
        "20": "fan failure",
        "21": "compressor failure",
    },
    "95": {  # Home Health Alarm
        "1": "leaving bed",
        "2": "sitting on bed",
        "3": "lying on bed",
        "4": "posture changed",
        "5": "sitting on edge of bed",
    },
    "96": {  # VOC Level
        "1": "clean",
        "2": "slightly polluted",
        "3": "moderately polluted",
        "4": "highly polluted",
    },
    "97": {  # Barrier Status
        **{
            "0": "closed",
            "100": "open",
            "101": "unknown",
            "102": "stopped",
            "103": "closing",
            "104": "opening",
        },
        **{
            str(b): f"{b} %" for a, b in enumerate(list(range(1, 100)))
        },  # 1-99 are percentage open
    },
    "98": {  # Insteon Thermostat Mode
        "0": "off",
        "1": "heat",
        "2": "cool",
        "3": "auto",
        "4": "fan_only",
        "5": "program_auto",
        "6": "program_heat",
        "7": "program_cool",
    },
    "99": {"7": "on", "8": "auto"},  # Insteon Thermostat Fan Mode
    "115": {  # Most recent On style action taken for lamp control
        "0": "on",
        "1": "off",
        "2": "fade up",
        "3": "fade down",
        "4": "fade stop",
        "5": "fast on",
        "6": "fast off",
        "7": "triple press on",
        "8": "triple press off",
        "9": "4x press on",
        "10": "4x press off",
        "11": "5x press on",
        "12": "5x press off",
    },
}

# Translate the "RR" Property to Seconds
INSTEON_RAMP_RATES = {
    "0": 540,
    "1": 480,
    "2": 420,
    "3": 360,
    "4": 300,
    "5": 270,
    "6": 240,
    "7": 210,
    "8": 180,
    "9": 150,
    "10": 120,
    "11": 90,
    "12": 60,
    "13": 47,
    "14": 43,
    "15": 38.5,
    "16": 34,
    "17": 32,
    "18": 30,
    "19": 28,
    "20": 26,
    "21": 23.5,
    "22": 21.5,
    "23": 19,
    "24": 8.5,
    "25": 6.5,
    "26": 4.5,
    "27": 2,
    "28": 0.5,
    "29": 0.3,
    "30": 0.2,
    "31": 0.1,
}

# Thermostat Types/Categories. 4.8 Trane, 5.3 venstar, 5.10 Insteon Wireless,
#  5.0x0B, 0x0F, 0x10, 0x13, 0x14, 0x15 - Insteon (alt. frequencies)
INSTEON_TYPE_THERMOSTAT = [
    "4.8",
    "5.3",
    "5.10",
    "5.11",
    "5.14",
    "5.15",
    "5.16",
    "5.17",
    "5.18",
    "5.19",
    "5.20",
    "5.21",
]
ZWAVE_CAT_THERMOSTAT = ["140"]

# Other special categories or types
INSTEON_TYPE_LOCK = ["4.64"]
ZWAVE_CAT_LOCK = ["111"]

INSTEON_TYPE_DIMMABLE = ["1."]
INSTEON_SUBNODE_DIMMABLE = " 1"
ZWAVE_CAT_DIMMABLE = ["109", "119", "186"]

# Insteon Battery Devices - States are ignored when checking the status of a group.
INSTEON_STATELESS_TYPE = ["0.16.", "0.17.", "0.18.", "16."]  # Not Used
INSTEON_STATELESS_NODEDEFID = [
    "BinaryAlarm",
    "BinaryAlarm_ADV",
    "BinaryControl",
    "BinaryControl_ADV",
    "RemoteLinc2",
    "RemoteLinc2_ADV",
    "DimmerSwitchOnly",
]

# Referenced from ISY-WSDK 4_fam.xml
# Included for user translations in external modules.
# This is the Node.zwave_props.category property.
DEVTYPE_CATEGORIES = {
    "0": "uninitialized",
    "101": "unknown",
    "102": "alarm",
    "103": "av control point",
    "104": "binary sensor",
    "105": "class a motor control",
    "106": "class b motor control",
    "107": "class c motor control",
    "108": "controller",
    "109": "dimmer switch",
    "110": "display",
    "111": "door lock",
    "112": "doorbell",
    "113": "entry control",
    "114": "gateway",
    "115": "installer tool",
    "116": "motor multiposition",
    "117": "climate sensor",
    "118": "multilevel sensor",
    "119": "multilevel switch",
    "120": "on/off power strip",
    "121": "on/off power switch",
    "122": "on/off scene switch",
    "123": "open/close valve",
    "124": "pc controller",
    "125": "remote",
    "126": "remote control",
    "127": "av remote control",
    "128": "simple remote control",
    "129": "repeater",
    "130": "residential hrv",
    "131": "satellite receiver",
    "132": "satellite receiver",
    "133": "scene controller",
    "134": "scene switch",
    "135": "security panel",
    "136": "set-top box",
    "137": "siren",
    "138": "smoke alarm",
    "139": "subsystem controller",
    "140": "thermostat",
    "141": "toggle",
    "142": "television",
    "143": "energy meter",
    "144": "pulse meter",
    "145": "water meter",
    "146": "gas meter",
    "147": "binary switch",
    "148": "binary alarm",
    "149": "aux alarm",
    "150": "co2 alarm",
    "151": "co alarm",
    "152": "freeze alarm",
    "153": "glass break alarm",
    "154": "heat alarm",
    "155": "motion sensor",
    "156": "smoke alarm",
    "157": "tamper alarm",
    "158": "tilt alarm",
    "159": "water alarm",
    "160": "door/window alarm",
    "161": "test alarm",
    "162": "low battery alarm",
    "163": "co end of life alarm",
    "164": "malfunction alarm",
    "165": "heartbeat",
    "166": "overheat alarm",
    "167": "rapid temp rise alarm",
    "168": "underheat alarm",
    "169": "leak detected alarm",
    "170": "level drop alarm",
    "171": "replace filter alarm",
    "172": "intrusion alarm",
    "173": "tamper code alarm",
    "174": "hardware failure alarm",
    "175": "software failure alarm",
    "176": "contact police alarm",
    "177": "contact fire alarm",
    "178": "contact medical alarm",
    "179": "wakeup alarm",
    "180": "timer",
    "181": "power management",
    "182": "appliance",
    "183": "home health",
    "184": "barrier",
    "185": "notification sensor",
    "186": "color switch",
    "187": "multilevel switch off on",
    "188": "multilevel switch down up",
    "189": "multilevel switch close open",
    "190": "multilevel switch ccw cw",
    "191": "multilevel switch left right",
    "192": "multilevel switch reverse forward",
    "193": "multilevel switch pull push",
    "194": "basic set",
    "195": "wall controller",
    "196": "barrier handle",
    "197": "sound switch",
}

# Referenced from ISY-WSDK cat.xml
# Included for user translations in external modules.
# This is the first part of the Node.type property (before the first ".")
NODE_CATEGORIES = {
    "0": "generic controller",
    "1": "dimming control",
    "2": "switch/relay control",
    "3": "network bridge",
    "4": "irrigation control",
    "5": "climate control",
    "6": "pool control",
    "7": "sensors/actuators",
    "8": "home entertainment",
    "9": "energy management",
    "10": "appliance control",
    "11": "plumbing",
    "12": "communications",
    "13": "computer",
    "14": "windows/shades",
    "15": "access control",
    "16": "security/health/safety",
    "17": "surveillance",
    "18": "automotive",
    "19": "pet care",
    "20": "toys",
    "21": "timers/clocks",
    "22": "holiday",
    "113": "a10/x10",
    "127": "virtual",
    "254": "unknown",
}

# Node Change Actions
NC_CLEAR_ERROR = "CE"
NC_FOLDER_ADDED = "FD"
NC_FOLDER_REMOVED = "FR"
NC_FOLDER_RENAMED = "FN"
NC_GROUP_ADDED = "GD"
NC_GROUP_REMOVED = "GR"
NC_GROUP_RENAMED = "GN"
NC_NET_RENAMED = "WR"
NC_NODE_ADDED = "ND"
NC_NODE_ENABLED = "EN"
NC_NODE_ERROR = "NE"
NC_NODE_MOVED = "MV"
NC_NODE_REMOVE_FROM_GROUP = "RG"
NC_NODE_REMOVED = "NR"
NC_NODE_RENAMED = "NN"
NC_NODE_REVISED = "RV"
NC_PARENT_CHANGED = "PC"
NC_PENDING_DEVICE_OP = "WH"
NC_PROGRAMMING_DEVICE = "WD"
DEV_WRITING = "_7A"
DEV_MEMORY = "_7M"

# Node Change Code: (Description, EventInfo Tags)
NODE_CHANGED_ACTIONS = {
    NC_CLEAR_ERROR: ("Node Comm. Errors Cleared", []),
    NC_FOLDER_ADDED: ("Folder Added", []),
    NC_FOLDER_REMOVED: ("Folder Removed", []),
    NC_FOLDER_RENAMED: ("Folder Renamed", ["newName"]),
    NC_GROUP_ADDED: ("Group Added", ["groupName", "groupType"]),
    NC_GROUP_REMOVED: ("Group Removed", []),
    NC_GROUP_RENAMED: ("Group Renamed", ["newName"]),
    NC_NET_RENAMED: ("Network Renamed", []),
    NC_NODE_ADDED: ("Node Added", ["nodeName", "nodeType"]),
    NC_NODE_ENABLED: ("Enabled/Disabled", ["enabled"]),
    NC_NODE_ERROR: ("Node Comm. Errors", []),
    NC_NODE_MOVED: ("Node moved into a Scene", ["movedNode", "linkType"]),
    NC_NODE_REMOVE_FROM_GROUP: ("Removed from Group (Scene)", ["removedNode"]),
    NC_NODE_REMOVED: ("Node Removed", []),
    NC_NODE_RENAMED: ("Node Renamed", ["newName"]),
    NC_NODE_REVISED: ("Node Revised (UPB)", []),
    NC_PARENT_CHANGED: ("Parent Changed", ["node", "nodeType", "parent", "parentType"]),
    NC_PENDING_DEVICE_OP: ("Pending Device Operation", []),
    NC_PROGRAMMING_DEVICE: ("Programming Device", []),
    DEV_WRITING: ("Progress Report", ["message"]),
    DEV_MEMORY: ("Memory Write", ["memory", "cmd1", "cmd2", "value"]),
}

SYSTEM_NOT_BUSY = "0"
SYSTEM_BUSY = "1"
SYSTEM_IDLE = "2"
SYSTEM_SAFE_MODE = "3"

SYSTEM_STATUS = {
    SYSTEM_NOT_BUSY: "Not Busy",
    SYSTEM_BUSY: "Busy",
    SYSTEM_IDLE: "Idle",
    SYSTEM_SAFE_MODE: "Safe Mode",
}

# Node Link Types
NODE_IS_CONTROLLER = 0x10

# Node operations flags
NODE_IS_INIT = 0x01  # needs to be initialized
NODE_TO_SCAN = 0x02  # needs to be scanned
NODE_IS_A_GROUP = 0x04  # it’s a group!
NODE_IS_ROOT = 0x08  # it’s the root group
NODE_IS_IN_ERR = 0x10  # it’s in error!
NODE_IS_NEW = 0x20  # brand new node
NODE_TO_DELETE = 0x40  # has to be deleted later
NODE_IS_DEVICE_ROOT = 0x80  # root device such as KPL load

DEV_CMD_MEMORY_WRITE = "0x2E"
DEV_BL_ADDR = "0x0264"
DEV_OL_ADDR = "0x0032"
DEV_RR_ADDR = "0x0021"

BACKLIGHT_SUPPORT = {
    "DimmerMotorSwitch": UOM_PERCENTAGE,
    "DimmerMotorSwitch_ADV": UOM_PERCENTAGE,
    "DimmerLampSwitch": UOM_PERCENTAGE,
    "DimmerLampSwitch_ADV": UOM_PERCENTAGE,
    "DimmerSwitchOnly": UOM_PERCENTAGE,
    "DimmerSwitchOnly_ADV": UOM_PERCENTAGE,
    "KeypadDimmer": UOM_INDEX,
    "KeypadDimmer_ADV": UOM_INDEX,
    "RelayLampSwitch": UOM_PERCENTAGE,
    "RelayLampSwitch_ADV": UOM_PERCENTAGE,
    "RelaySwitchOnlyPlusQuery": UOM_PERCENTAGE,
    "RelaySwitchOnlyPlusQuery_ADV": UOM_PERCENTAGE,
    "RelaySwitchOnly": UOM_PERCENTAGE,
    "RelaySwitchOnly_ADV": UOM_PERCENTAGE,
    "KeypadRelay": UOM_INDEX,
    "KeypadRelay_ADV": UOM_INDEX,
    "KeypadButton": UOM_INDEX,
    "KeypadButton_ADV": UOM_INDEX,
}

BACKLIGHT_INDEX = [
    "On  0 / Off 0",
    "On  1 / Off 0",
    "On  2 / Off 0",
    "On  3 / Off 0",
    "On  4 / Off 0",
    "On  5 / Off 0",
    "On  6 / Off 0",
    "On  7 / Off 0",
    "On  8 / Off 0",
    "On  9 / Off 0",
    "On 10 / Off 0",
    "On 11 / Off 0",
    "On 12 / Off 0",
    "On 13 / Off 0",
    "On 14 / Off 0",
    "On 15 / Off 0",
    "On  0 / Off 1",
    "On  1 / Off 1",
    "On  2 / Off 1",
    "On  3 / Off 1",
    "On  4 / Off 1",
    "On  5 / Off 1",
    "On  6 / Off 1",
    "On  7 / Off 1",
    "On  8 / Off 1",
    "On  9 / Off 1",
    "On 10 / Off 1",
    "On 11 / Off 1",
    "On 12 / Off 1",
    "On 13 / Off 1",
    "On 14 / Off 1",
    "On 15 / Off 1",
    "On  0 / Off 2",
    "On  1 / Off 2",
    "On  2 / Off 2",
    "On  3 / Off 2",
    "On  4 / Off 2",
    "On  5 / Off 2",
    "On  6 / Off 2",
    "On  7 / Off 2",
    "On  8 / Off 2",
    "On  9 / Off 2",
    "On 10 / Off 2",
    "On 11 / Off 2",
    "On 12 / Off 2",
    "On 13 / Off 2",
    "On 14 / Off 2",
    "On 15 / Off 2",
    "On  0 / Off 3",
    "On  1 / Off 3",
    "On  2 / Off 3",
    "On  3 / Off 3",
    "On  4 / Off 3",
    "On  5 / Off 3",
    "On  6 / Off 3",
    "On  7 / Off 3",
    "On  8 / Off 3",
    "On  9 / Off 3",
    "On 10 / Off 3",
    "On 11 / Off 3",
    "On 12 / Off 3",
    "On 13 / Off 3",
    "On 14 / Off 3",
    "On 15 / Off 3",
    "On  0 / Off 4",
    "On  1 / Off 4",
    "On  2 / Off 4",
    "On  3 / Off 4",
    "On  4 / Off 4",
    "On  5 / Off 4",
    "On  6 / Off 4",
    "On  7 / Off 4",
    "On  8 / Off 4",
    "On  9 / Off 4",
    "On 10 / Off 4",
    "On 11 / Off 4",
    "On 12 / Off 4",
    "On 13 / Off 4",
    "On 14 / Off 4",
    "On 15 / Off 4",
    "On  0 / Off 5",
    "On  1 / Off 5",
    "On  2 / Off 5",
    "On  3 / Off 5",
    "On  4 / Off 5",
    "On  5 / Off 5",
    "On  6 / Off 5",
    "On  7 / Off 5",
    "On  8 / Off 5",
    "On  9 / Off 5",
    "On 10 / Off 5",
    "On 11 / Off 5",
    "On 12 / Off 5",
    "On 13 / Off 5",
    "On 14 / Off 5",
    "On 15 / Off 5",
    "On  0 / Off 6",
    "On  1 / Off 6",
    "On  2 / Off 6",
    "On  3 / Off 6",
    "On  4 / Off 6",
    "On  5 / Off 6",
    "On  6 / Off 6",
    "On  7 / Off 6",
    "On  8 / Off 6",
    "On  9 / Off 6",
    "On 10 / Off 6",
    "On 11 / Off 6",
    "On 12 / Off 6",
    "On 13 / Off 6",
    "On 14 / Off 6",
    "On 15 / Off 6",
    "On  0 / Off 7",
    "On  1 / Off 7",
    "On  2 / Off 7",
    "On  3 / Off 7",
    "On  4 / Off 7",
    "On  5 / Off 7",
    "On  6 / Off 7",
    "On  7 / Off 7",
    "On  8 / Off 7",
    "On  9 / Off 7",
    "On 10 / Off 7",
    "On 11 / Off 7",
    "On 12 / Off 7",
    "On 13 / Off 7",
    "On 14 / Off 7",
    "On 15 / Off 7",
]



"""ISY Event Stream."""
import asyncio
import logging
import socket
import ssl
from threading import Thread, ThreadError
import time
import xml
from xml.dom import minidom

from . import strings
from ..constants import (
    ACTION_KEY,
    ACTION_KEY_CHANGED,
    ATTR_ACTION,
    ATTR_CONTROL,
    ATTR_ID,
    ATTR_STREAM_ID,
    ATTR_VAR,
    ES_CONNECTED,
    ES_DISCONNECTED,
    ES_INITIALIZING,
    ES_LOADED,
    ES_LOST_STREAM_CONNECTION,
    POLL_TIME,
    PROP_STATUS,
    RECONNECT_DELAY,
    TAG_EVENT_INFO,
    TAG_NODE,
)
from ..exceptions import ISYInvalidAuthError, ISYMaxConnections, ISYStreamDataError
from ..helpers import attr_from_xml, now, value_from_xml
from ..logging import LOG_VERBOSE
from .eventreader import ISYEventReader

_LOGGER = logging.getLogger(__name__)  # Allows targeting pyisy.events in handlers.


class EventStream:
    """Class to represent the Event Stream from the ISY."""

    def __init__(self, isy, connection_info, on_lost_func=None):
        """Initialize the EventStream class."""
        self.isy = isy
        self._running = False
        self._writer = None
        self._thread = None
        self._subscribed = False
        self._connected = False
        self._lasthb = None
        self._hbwait = 0
        self._loaded = None
        self._on_lost_function = on_lost_func
        self._program_key = None
        self.cert = None
        self.data = connection_info

        # create TLS encrypted socket if we're using HTTPS
        if self.data.get("tls"):
            if self.data["tls"] == 1.1:
                context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_1)
            else:
                context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
            context.check_hostname = False
            self.socket = context.wrap_socket(
                socket.socket(socket.AF_INET, socket.SOCK_STREAM),
                server_hostname=f"https://{self.data['addr']}",
            )
        else:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def _create_message(self, msg):
        """Prepare a message for sending."""
        head = msg["head"]
        body = msg["body"]
        body = body.format(**self.data)
        length = len(body)
        head = head.format(length=length, **self.data)
        return head + body

    def _route_message(self, msg):
        """Route a received message from the event stream."""
        # check xml formatting
        try:
            xmldoc = minidom.parseString(msg)
        except xml.parsers.expat.ExpatError:
            _LOGGER.warning("ISY Received Malformed XML:\n%s", msg)
            return
        _LOGGER.log(LOG_VERBOSE, "ISY Update Received:\n%s", msg)

        # A wild stream id appears!
        if f"{ATTR_STREAM_ID}=" in msg and ATTR_STREAM_ID not in self.data:
            self.update_received(xmldoc)

        # direct the event message
        cntrl = value_from_xml(xmldoc, ATTR_CONTROL)
        if not cntrl:
            return
        if cntrl == "_0":  # ISY HEARTBEAT
            if self._loaded is None:
                self._loaded = ES_INITIALIZING
                self.isy.connection_events.notify(ES_INITIALIZING)
            elif self._loaded == ES_INITIALIZING:
                self._loaded = ES_LOADED
                self.isy.connection_events.notify(ES_LOADED)
            self._lasthb = now()
            self._hbwait = int(value_from_xml(xmldoc, ATTR_ACTION))
            _LOGGER.debug("ISY HEARTBEAT: %s", self._lasthb.isoformat())
        elif cntrl == PROP_STATUS:  # NODE UPDATE
            self.isy.nodes.update_received(xmldoc)
        elif cntrl[0] != "_":  # NODE CONTROL EVENT
            self.isy.nodes.control_message_received(xmldoc)
        elif cntrl == "_1":  # Trigger Update
            if f"<{ATTR_VAR}" in msg:  # VARIABLE
                self.isy.variables.update_received(xmldoc)
            elif f"<{ATTR_ID}>" in msg:  # PROGRAM
                self.isy.programs.update_received(xmldoc)
            elif f"<{TAG_NODE}>" in msg and "[" in msg:  # Node Server Update
                pass  # This is most likely a duplicate node update.
            elif f"<{ATTR_ACTION}>" in msg:
                action = value_from_xml(xmldoc, ATTR_ACTION)
                if action == ACTION_KEY:
                    self.data[ACTION_KEY] = value_from_xml(xmldoc, TAG_EVENT_INFO)
                    return
                if action == ACTION_KEY_CHANGED:
                    self._program_key = value_from_xml(xmldoc, TAG_NODE)
                # Need to reload programs
                asyncio.run_coroutine_threadsafe(
                    self.isy.programs.update(), self.isy.loop
                )
        elif cntrl == "_3":  # Node Changed/Updated
            self.isy.nodes.node_changed_received(xmldoc)

    def update_received(self, xmldoc):
        """Set the socket ID."""
        self.data[ATTR_STREAM_ID] = attr_from_xml(xmldoc, "Event", ATTR_STREAM_ID)
        _LOGGER.debug("ISY Updated Events Stream ID %s", self.data[ATTR_STREAM_ID])

    @property
    def running(self):
        """Return the running state of the thread."""
        try:
            return self._thread.isAlive()
        except (AttributeError, RuntimeError, ThreadError):
            return False

    @running.setter
    def running(self, val):
        if val and not self.running:
            _LOGGER.info("ISY Starting Updates")
            if self.connect():
                self.subscribe()
                self._running = True
                self._thread = Thread(target=self.watch)
                self._thread.daemon = True
                self._thread.start()
        else:
            _LOGGER.info("ISY Stopping Updates")
            self._running = False
            self.unsubscribe()
            self.disconnect()

    def write(self, msg):
        """Write data back to the socket."""
        if self._writer is None:
            raise NotImplementedError("Function not available while socket is closed.")
        self._writer.write(msg)
        self._writer.flush()

    def connect(self):
        """Connect to the event stream socket."""
        if not self._connected:
            try:
                self.socket.connect((self.data["addr"], self.data["port"]))
                if self.data.get("tls"):
                    self.cert = self.socket.getpeercert()
            except OSError as err:
                _LOGGER.exception(
                    "PyISY could not connect to ISY event stream. %s", err
                )
                if self._on_lost_function is not None:
                    self._on_lost_function()
                return False
            self.socket.setblocking(0)
            self._writer = self.socket.makefile("w")
            self._connected = True
            self.isy.connection_events.notify(ES_CONNECTED)
            return True
        return True

    def disconnect(self):
        """Disconnect from the Event Stream socket."""
        if self._connected:
            self.socket.close()
            self._connected = False
            self._subscribed = False
            self._running = False
            self.isy.connection_events.notify(ES_DISCONNECTED)

    def subscribe(self):
        """Subscribe to the Event Stream."""
        if not self._subscribed and self._connected:
            if ATTR_STREAM_ID not in self.data:
                msg = self._create_message(strings.SUB_MSG)
                self.write(msg)
            else:
                msg = self._create_message(strings.RESUB_MSG)
                self.write(msg)
            self._subscribed = True

    def unsubscribe(self):
        """Unsubscribe from the Event Stream."""
        if self._subscribed and self._connected:
            msg = self._create_message(strings.UNSUB_MSG)
            try:
                self.write(msg)
            except OSError as ex:
                _LOGGER.error(
                    "PyISY encountered a socket error while writing unsubscribe message to the socket: %s.",
                    ex,
                )
            self._subscribed = False
            self.disconnect()

    @property
    def connected(self):
        """Return if the module is connected to the ISY or not."""
        return self._connected

    @property
    def heartbeat_time(self):
        """Return the last ISY Heartbeat time."""
        if self._lasthb is not None:
            return (now() - self._lasthb).seconds
        return 0.0

    def _lost_connection(self, delay=0):
        """React when the event stream connection is lost."""
        _LOGGER.warning("PyISY lost connection to the ISY event stream.")
        self.isy.connection_events.notify(ES_LOST_STREAM_CONNECTION)
        self.unsubscribe()
        if self._on_lost_function is not None:
            time.sleep(delay)
            self._on_lost_function()

    def watch(self):
        """Watch the subscription connection and report if dead."""
        if not self._subscribed:
            _LOGGER.debug("PyISY watch called without a subscription.")
            return

        event_reader = ISYEventReader(self.socket)

        while self._running and self._subscribed:
            # verify connection is still alive
            if self.heartbeat_time > self._hbwait:
                self._lost_connection()
                return

            try:
                events = event_reader.read_events(POLL_TIME)
            except ISYMaxConnections:
                _LOGGER.error(
                    "PyISY reached maximum connections, delaying reconnect attempt by %s seconds.",
                    RECONNECT_DELAY,
                )
                self._lost_connection(RECONNECT_DELAY)
                return
            except ISYInvalidAuthError:
                _LOGGER.error(
                    "Invalid authentication used to connect to the event stream."
                )
                return
            except ISYStreamDataError as ex:
                _LOGGER.warning(
                    "PyISY encountered an error while reading the event stream: %s.", ex
                )
                self._lost_connection()
                return
            except OSError as ex:
                _LOGGER.warning(
                    "PyISY encountered a socket error while reading the event stream: %s.",
                    ex,
                )
                self._lost_connection()
                return

            for message in events:
                try:
                    self._route_message(message)
                except Exception as ex:  # pylint: disable=broad-except
                    _LOGGER.warning(
                        "PyISY encountered while routing message '%s': %s", message, ex
                    )
                    raise

    def __del__(self):
        """Ensure we unsubscribe on destroy."""
        self.unsubscribe()



"""ISY TCP Socket Event Reader."""
import errno
import select
import ssl

from ..constants import SOCKET_BUFFER_SIZE
from ..exceptions import (
    ISYInvalidAuthError,
    ISYMaxConnections,
    ISYStreamDataError,
    ISYStreamDisconnected,
)


class ISYEventReader:
    """Read in streams of ISY HTTP Events."""

    HTTP_HEADER_SEPERATOR = b"\r\n"
    HTTP_HEADER_BODY_SEPERATOR = b"\r\n\r\n"
    HTTP_HEADER_BODY_SEPERATOR_LEN = 4
    REACHED_MAX_CONNECTIONS_RESPONSE = b"HTTP/1.1 817"
    HTTP_NOT_AUTHORIZED_RESPONSE = b"HTTP/1.1 401"
    CONTENT_LENGTH_HEADER = b"content-length"
    HEADER_SEPERATOR = b":"

    def __init__(self, isy_read_socket):
        """Initialize the ISYEventStream class."""
        self._event_buffer = b""
        self._event_content_length = None
        self._event_count = 0
        self._socket = isy_read_socket

    def read_events(self, timeout):
        """Read events from the socket."""
        events = []
        # poll socket for new data
        if not self._receive_into_buffer(timeout):
            return events

        while True:
            # Read the headers if we do not have content length yet
            if not self._event_content_length:
                seperator_position = self._event_buffer.find(
                    self.HTTP_HEADER_BODY_SEPERATOR
                )
                if seperator_position == -1:
                    return events
                self._parse_headers(seperator_position)

            # If we do not have a body yet
            if len(self._event_buffer) < self._event_content_length:
                return events

            # We have the body now
            body = self._event_buffer[0 : self._event_content_length]
            self._event_count += 1
            self._event_buffer = self._event_buffer[self._event_content_length :]
            self._event_content_length = None
            events.append(body.decode(encoding="utf-8", errors="ignore"))

    def _receive_into_buffer(self, timeout):
        """Receive data on available on the socket.

        If we get an empty read on the first read attempt
        this means the isy has disconnected.

        If we get an empty read on the first read attempt
        and we have seen only one event, the isy has reached
        the maximum number of event listeners.
        """
        inready, _, _ = select.select([self._socket], [], [], timeout)
        if self._socket not in inready:
            return False

        try:
            # We have data on the wire, read as much as we can
            # up to 32 * SOCKET_BUFFER_SIZE
            for read_count in range(0, 32):
                new_data = self._socket.recv(SOCKET_BUFFER_SIZE)
                print(f"read_count: {read_count} new_data: {new_data}")
                if len(new_data) == 0:
                    if read_count != 0:
                        break
                    if self._event_count <= 1:
                        raise ISYMaxConnections(self._event_buffer)
                    raise ISYStreamDisconnected(self._event_buffer)

                self._event_buffer += new_data
        except ssl.SSLWantReadError:
            pass
        except OSError as ex:
            if ex.errno != errno.EWOULDBLOCK:
                raise

        return True

    def _parse_headers(self, seperator_position):
        """Find the content-length in the headers."""
        headers = self._event_buffer[0:seperator_position]
        if headers.startswith(self.REACHED_MAX_CONNECTIONS_RESPONSE):
            raise ISYMaxConnections(self._event_buffer)
        if headers.startswith(self.HTTP_NOT_AUTHORIZED_RESPONSE):
            raise ISYInvalidAuthError(self._event_buffer)
        self._event_buffer = self._event_buffer[
            seperator_position + self.HTTP_HEADER_BODY_SEPERATOR_LEN :
        ]
        for header in headers.split(self.HTTP_HEADER_SEPERATOR)[1:]:
            header_name, header_value = header.split(self.HEADER_SEPERATOR, 1)
            if header_name.strip().lower() != self.CONTENT_LENGTH_HEADER:
                continue
            self._event_content_length = int(header_value.strip())
        if not self._event_content_length:
            raise ISYStreamDataError(headers)



"""
PyISY - Python Library for the ISY Controller.

This module is a set of Python bindings for the ISY's REST API. The
ISY is developed by Universal Devices and is a home automation
controller for Insteon and X10 devices.

Copyright 2015 Ryan M. Kraus
               rmkraus at gmail dot com

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from importlib.metadata import PackageNotFoundError, version

from .exceptions import (
    ISYConnectionError,
    ISYInvalidAuthError,
    ISYMaxConnections,
    ISYResponseParseError,
    ISYStreamDataError,
    ISYStreamDisconnected,
)
from .isy import ISY

try:
    __version__ = version("pyisy")
except PackageNotFoundError:
    __version__ = "unknown"

__all__ = [
    "ISY",
    "ISYInvalidAuthError",
    "ISYConnectionError",
    "ISYMaxConnections",
    "ISYResponseParseError",
    "ISYStreamDataError",
    "ISYStreamDisconnected",
]
__author__ = "Ryan M. Kraus"
__email__ = "rmkraus at gmail dot com"
__date__ = "February 2020"



"""Base object for nodes and groups."""
from xml.dom import minidom

from ..constants import (
    ATTR_LAST_CHANGED,
    ATTR_LAST_UPDATE,
    ATTR_STATUS,
    CMD_BEEP,
    CMD_BRIGHTEN,
    CMD_DIM,
    CMD_DISABLE,
    CMD_ENABLE,
    CMD_FADE_DOWN,
    CMD_FADE_STOP,
    CMD_FADE_UP,
    CMD_OFF,
    CMD_OFF_FAST,
    CMD_ON,
    CMD_ON_FAST,
    COMMAND_FRIENDLY_NAME,
    METHOD_COMMAND,
    NODE_FAMILY_ID,
    TAG_ADDRESS,
    TAG_DESCRIPTION,
    TAG_IS_LOAD,
    TAG_LOCATION,
    TAG_NAME,
    TAG_SPOKEN,
    URL_CHANGE,
    URL_NODES,
    URL_NOTES,
    XML_TRUE,
)
from ..exceptions import XML_ERRORS, XML_PARSE_ERROR, ISYResponseParseError
from ..helpers import EventEmitter, NodeProperty, now, value_from_xml
from ..logging import _LOGGER


class NodeBase:
    """Base Object for Nodes and Groups/Scenes."""

    has_children = False

    def __init__(
        self,
        nodes,
        address,
        name,
        status,
        family_id=None,
        aux_properties=None,
        pnode=None,
        flag=0,
    ):
        """Initialize a Node Base class."""
        self._aux_properties = aux_properties if aux_properties is not None else {}
        self._family = NODE_FAMILY_ID.get(family_id)
        self._id = address
        self._name = name
        self._nodes = nodes
        self._notes = None
        self._primary_node = pnode
        self._flag = flag
        self._status = status
        self._last_update = now()
        self._last_changed = now()
        self.isy = nodes.isy
        self.status_events = EventEmitter()

    def __str__(self):
        """Return a string representation of the node."""
        return f"{type(self).__name__}({self._id})"

    @property
    def aux_properties(self):
        """Return the aux properties that were in the Node Definition."""
        return self._aux_properties

    @property
    def address(self):
        """Return the Node ID."""
        return self._id

    @property
    def description(self):
        """Return the description of the node from it's notes."""
        if self._notes is None:
            _LOGGER.debug(
                "No notes retrieved for node. Call get_notes() before accessing."
            )
        return self._notes[TAG_DESCRIPTION]

    @property
    def family(self):
        """Return the ISY Family category."""
        return self._family

    @property
    def flag(self):
        """Return the flag of the current node as a property."""
        return self._flag

    @property
    def folder(self):
        """Return the folder of the current node as a property."""
        return self._nodes.get_folder(self.address)

    @property
    def is_load(self):
        """Return the isLoad property of the node from it's notes."""
        if self._notes is None:
            _LOGGER.debug(
                "No notes retrieved for node. Call get_notes() before accessing."
            )
        return self._notes[TAG_IS_LOAD]

    @property
    def last_changed(self):
        """Return the UTC Time of the last status change for this node."""
        return self._last_changed

    @property
    def last_update(self):
        """Return the UTC Time of the last update for this node."""
        return self._last_update

    @property
    def location(self):
        """Return the location of the node from it's notes."""
        if self._notes is None:
            _LOGGER.debug(
                "No notes retrieved for node. Call get_notes() before accessing."
            )
        return self._notes[TAG_LOCATION]

    @property
    def name(self):
        """Return the name of the Node."""
        return self._name

    @property
    def primary_node(self):
        """Return just the parent/primary node address.

        This is similar to Node.parent_node but does not return the whole Node
        class, and will return itself if it is the primary node/group.

        """
        return self._primary_node

    @property
    def spoken(self):
        """Return the text of the Spoken property inside the group notes."""
        if self._notes is None:
            _LOGGER.debug(
                "No notes retrieved for node. Call get_notes() before accessing."
            )
        return self._notes[TAG_SPOKEN]

    @property
    def status(self):
        """Return the current node state."""
        return self._status

    @status.setter
    def status(self, value):
        """Set the current node state and notify listeners."""
        if self._status != value:
            self._status = value
            self._last_changed = now()
            self.status_events.notify(self.status_feedback)
        return self._status

    @property
    def status_feedback(self):
        """Return information for a status change event."""
        return {
            TAG_ADDRESS: self.address,
            ATTR_STATUS: self._status,
            ATTR_LAST_CHANGED: self._last_changed,
            ATTR_LAST_UPDATE: self._last_update,
        }

    async def get_notes(self):
        """Retrieve and parse the notes for a given node.

        Notes are not retrieved unless explicitly requested by
        a call to this function.
        """
        notes_xml = await self.isy.conn.request(
            self.isy.conn.compile_url([URL_NODES, self._id, URL_NOTES]), ok404=True
        )
        spoken = None
        is_load = None
        description = None
        location = None
        if notes_xml is not None and notes_xml != "":
            try:
                notes_dom = minidom.parseString(notes_xml)
            except XML_ERRORS as exc:
                _LOGGER.error("%s: Node Notes %s", XML_PARSE_ERROR, notes_xml)
                raise ISYResponseParseError() from exc

            spoken = value_from_xml(notes_dom, TAG_SPOKEN)
            location = value_from_xml(notes_dom, TAG_LOCATION)
            description = value_from_xml(notes_dom, TAG_DESCRIPTION)
            is_load = value_from_xml(notes_dom, TAG_IS_LOAD)
        return {
            TAG_SPOKEN: spoken,
            TAG_IS_LOAD: is_load == XML_TRUE,
            TAG_DESCRIPTION: description,
            TAG_LOCATION: location,
        }

    async def update(self, event=None, wait_time=0, xmldoc=None):
        """Update the group with values from the controller."""
        self.update_last_update()

    def update_property(self, prop):
        """Update an aux property for the node when received."""
        if not isinstance(prop, NodeProperty):
            _LOGGER.error("Could not update property value. Invalid type provided.")
            return
        self.update_last_update()

        aux_prop = self.aux_properties.get(prop.control)
        if aux_prop:
            if prop.uom == "" and not aux_prop.uom == "":
                # Guard against overwriting known UOM with blank UOM (ISYv4).
                prop.uom = aux_prop.uom
            if aux_prop == prop:
                return
        self.aux_properties[prop.control] = prop
        self.update_last_changed()
        self.status_events.notify(self.status_feedback)

    def update_last_changed(self, timestamp=None):
        """Set the UTC Time of the last status change for this node."""
        if timestamp is None:
            timestamp = now()
        self._last_changed = timestamp

    def update_last_update(self, timestamp=None):
        """Set the UTC Time of the last update for this node."""
        if timestamp is None:
            timestamp = now()
        self._last_update = timestamp

    async def send_cmd(self, cmd, val=None, uom=None, query=None):
        """Send a command to the device."""
        value = str(val) if val is not None else None
        _uom = str(uom) if uom is not None else None
        req = [URL_NODES, str(self._id), METHOD_COMMAND, cmd]
        if value:
            req.append(value)
        if _uom:
            req.append(_uom)
        req_url = self.isy.conn.compile_url(req, query)
        if not await self.isy.conn.request(req_url):
            _LOGGER.warning(
                "ISY could not send %s command to %s.",
                COMMAND_FRIENDLY_NAME.get(cmd),
                self._id,
            )
            return False
        _LOGGER.debug(
            "ISY command %s sent to %s.", COMMAND_FRIENDLY_NAME.get(cmd), self._id
        )
        return True

    async def beep(self):
        """Identify physical device by sound (if supported)."""
        return await self.send_cmd(CMD_BEEP)

    async def brighten(self):
        """Increase brightness of a device by ~3%."""
        return await self.send_cmd(CMD_BRIGHTEN)

    async def dim(self):
        """Decrease brightness of a device by ~3%."""
        return await self.send_cmd(CMD_DIM)

    async def disable(self):
        """Send command to the node to disable it."""
        if not await self.isy.conn.request(
            self.isy.conn.compile_url([URL_NODES, str(self._id), CMD_DISABLE])
        ):
            _LOGGER.warning("ISY could not %s %s.", CMD_DISABLE, self._id)
            return False
        return True

    async def enable(self):
        """Send command to the node to enable it."""
        if not await self.isy.conn.request(
            self.isy.conn.compile_url([URL_NODES, str(self._id), CMD_ENABLE])
        ):
            _LOGGER.warning("ISY could not %s %s.", CMD_ENABLE, self._id)
            return False
        return True

    async def fade_down(self):
        """Begin fading down (dim) a device."""
        return await self.send_cmd(CMD_FADE_DOWN)

    async def fade_stop(self):
        """Stop fading a device."""
        return await self.send_cmd(CMD_FADE_STOP)

    async def fade_up(self):
        """Begin fading up (dim) a device."""
        return await self.send_cmd(CMD_FADE_UP)

    async def fast_off(self):
        """Start manually brightening a device."""
        return await self.send_cmd(CMD_OFF_FAST)

    async def fast_on(self):
        """Start manually brightening a device."""
        return await self.send_cmd(CMD_ON_FAST)

    async def query(self):
        """Request the ISY query this node."""
        return await self.isy.query(address=self.address)

    async def turn_off(self):
        """Turn off the nodes/group in the ISY."""
        return await self.send_cmd(CMD_OFF)

    async def turn_on(self, val=None):
        """
        Turn the node on.

        |  [optional] val: The value brightness value (0-255) for the node.
        """
        if val is None or type(self).__name__ == "Group":
            cmd = CMD_ON
        elif int(val) > 0:
            cmd = CMD_ON
            val = str(val) if int(val) <= 255 else None
        else:
            cmd = CMD_OFF
            val = None
        return await self.send_cmd(cmd, val)

    async def rename(self, new_name):
        """
        Rename the node or group in the ISY.

        Note: Feature was added in ISY v5.2.0, this will fail on earlier versions.
        """

        # /rest/nodes/<nodeAddress>/change?name=<newName>
        req_url = self.isy.conn.compile_url(
            [URL_NODES, self._id, URL_CHANGE],
            query={TAG_NAME: new_name},
        )
        if not await self.isy.conn.request(req_url):
            _LOGGER.warning(
                "ISY could not update name for %s.",
                self._id,
            )
            return False
        _LOGGER.debug("ISY renamed %s to %s.", self._id, new_name)

        self._name = new_name
        return True



"""Module Setup File for PIP Installation."""

import pathlib

from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="pyisy",
    version_format="{tag}",
    license="Apache License 2.0",
    url="https://github.com/automicus/PyISY",
    author="Ryan Kraus",
    author_email="automicus@gmail.com",
    description="Python module to talk to ISY994 from UDI.",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    install_requires=[
        "aiohttp>=3.8.1",
        "python-dateutil>=2.8.1",
        "requests>=2.28.1",
        "colorlog>=6.6.0",
    ],
    keywords=["home automation", "isy", "isy994", "isy-994", "UDI"],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Home Automation",
    ],
)



"""Module for connecting to and interacting with the ISY."""
import asyncio
from threading import Thread

from .clock import Clock
from .configuration import Configuration
from .connection import Connection
from .constants import (
    CMD_X10,
    ES_CONNECTED,
    ES_RECONNECT_FAILED,
    ES_RECONNECTING,
    ES_START_UPDATES,
    ES_STOP_UPDATES,
    PROTO_ISY,
    URL_QUERY,
    X10_COMMANDS,
)
from .events.tcpsocket import EventStream
from .events.websocket import WebSocketClient
from .helpers import EventEmitter
from .logging import _LOGGER, enable_logging
from .networking import NetworkResources
from .nodes import Nodes
from .programs import Programs
from .variables import Variables


class ISY:
    """
    This is the main class that handles interaction with the ISY device.

    |  address: String of the IP address of the ISY device
    |  port: String of the port over which the ISY is serving its API
    |  username: String of the administrator username for the ISY
    |  password: String of the administrator password for the ISY
    |  use_https: [optional] Boolean of whether secured HTTP should be used
    |  tls_ver: [optional] Number indicating the version of TLS encryption to
       use. Valid options are 1.1 or 1.2.

    :ivar auto_reconnect: Boolean value that indicates if the class should
                          auto-reconnect to the event stream if the connection
                          is lost.
    :ivar auto_update: Boolean value that controls the class's subscription to
                       the event stream that allows node, program
                       values to be updated automatically.
    :ivar connected: Read only boolean value indicating if the class is
                     connected to the controller.
    :ivar nodes: :class:`pyisy.nodes.Nodes` manager that interacts with
                 Insteon nodes and groups.
    :ivar programs: Program manager that interacts with ISY programs and i
                    folders.
    :ivar variables: Variable manager that interacts with ISY variables.
    """

    auto_reconnect = True

    def __init__(
        self,
        address,
        port,
        username,
        password,
        use_https=False,
        tls_ver=1.1,
        webroot="",
        websession=None,
        use_websocket=False,
    ):
        """Initialize the primary ISY Class."""
        self._events = None  # create this JIT so no socket reuse
        self._reconnect_thread = None
        self._connected = False

        if len(_LOGGER.handlers) == 0:
            enable_logging(add_null_handler=True)

        self.conn = Connection(
            address=address,
            port=port,
            username=username,
            password=password,
            use_https=use_https,
            tls_ver=tls_ver,
            webroot=webroot,
            websession=websession,
        )

        self.websocket = None
        if use_websocket:
            self.websocket = WebSocketClient(
                isy=self,
                address=address,
                port=port,
                username=username,
                password=password,
                use_https=use_https,
                tls_ver=tls_ver,
                webroot=webroot,
                websession=websession,
            )

        self.configuration = None
        self.clock = None
        self.nodes = None
        self.node_servers = None
        self.programs = None
        self.variables = None
        self.networking = None
        self._hostname = address
        self.connection_events = EventEmitter()
        self.system_status = self.conn.system_status
        self.loop = asyncio.get_running_loop()
        self._uuid = None

    async def initialize(self, with_node_servers=False):
        """Initialize the connection with the ISY."""
        config_xml = await self.conn.test_connection()
        self.configuration = Configuration(xml=config_xml)
        self._uuid = self.configuration["uuid"]
        if not self.configuration["model"].startswith("ISY 994"):
            self.conn.increase_available_connections()

        isy_setup_tasks = [
            self.conn.get_status(),
            self.conn.get_time(),
            self.conn.get_nodes(),
            self.conn.get_programs(),
            self.conn.get_variable_defs(),
            self.conn.get_variables(),
        ]
        if self.configuration["Networking Module"]:
            isy_setup_tasks.append(asyncio.create_task(self.conn.get_network()))
        isy_setup_results = await asyncio.gather(*isy_setup_tasks)

        self.clock = Clock(self, xml=isy_setup_results[1])
        self.nodes = Nodes(self, xml=isy_setup_results[2])
        self.programs = Programs(self, xml=isy_setup_results[3])
        self.variables = Variables(
            self,
            def_xml=isy_setup_results[4],
            var_xml=isy_setup_results[5],
        )
        if self.configuration["Networking Module"]:
            self.networking = NetworkResources(self, xml=isy_setup_results[6])
        await self.nodes.update(xml=isy_setup_results[0])
        if self.node_servers and with_node_servers:
            await self.node_servers.load_node_servers()

        self._connected = True

    async def shutdown(self):
        """Cleanup connections and prepare for exit."""
        if self.websocket is not None:
            self.websocket.stop()
        if self._events is not None and self._events.running:
            self.connection_events.notify(ES_STOP_UPDATES)
            self._events.running = False
        await self.conn.close()

    @property
    def conf(self):
        """Return the status of the connection (shortcut property)."""
        return self.configuration

    @property
    def connected(self):
        """Return the status of the connection."""
        return self._connected

    @property
    def auto_update(self):
        """Return the auto_update property."""
        if self.websocket is not None:
            return self.websocket.status == ES_CONNECTED
        if self._events is not None:
            return self._events.running
        return False

    @auto_update.setter
    def auto_update(self, val):
        """Set the auto_update property."""
        if self.websocket is not None:
            _LOGGER.warning(
                "Websockets are enabled. Use isy.websocket.start() or .stop() instead."
            )
            return
        if val and not self.auto_update:
            # create new event stream socket
            self._events = EventStream(
                self, self.conn.connection_info, self._on_lost_event_stream
            )
        if self._events is not None:
            self.connection_events.notify(ES_START_UPDATES if val else ES_STOP_UPDATES)
            self._events.running = val

    @property
    def hostname(self):
        """Return the hostname."""
        return self._hostname

    @property
    def protocol(self):
        """Return the protocol for this entity."""
        return PROTO_ISY

    @property
    def uuid(self):
        """Return the ISY's uuid."""
        return self._uuid

    def _on_lost_event_stream(self):
        """Handle lost connection to event stream."""
        del self._events
        self._events = None

        if self.auto_reconnect and self._reconnect_thread is None:
            # attempt to reconnect
            self._reconnect_thread = Thread(target=self._auto_reconnecter)
            self._reconnect_thread.daemon = True
            self._reconnect_thread.start()

    def _auto_reconnecter(self):
        """Auto-reconnect to the event stream."""
        while self.auto_reconnect and not self.auto_update:
            _LOGGER.warning("PyISY attempting stream reconnect.")
            del self._events
            self._events = EventStream(
                self, self.conn.connection_info, self._on_lost_event_stream
            )
            self._events.running = True
            self.connection_events.notify(ES_RECONNECTING)

        if not self.auto_update:
            del self._events
            self._events = None
            _LOGGER.warning("PyISY could not reconnect to the event stream.")
            self.connection_events.notify(ES_RECONNECT_FAILED)
        else:
            _LOGGER.warning("PyISY reconnected to the event stream.")

        self._reconnect_thread = None

    async def query(self, address=None):
        """Query all the nodes or a specific node if an address is provided .

        Args:
            address (string, optional): Node Address to query. Defaults to None.

        Returns:
            boolean: Returns `True` on successful command, `False` on error.
        """
        req_path = [URL_QUERY]
        if address is not None:
            req_path.append(address)
        req_url = self.conn.compile_url(req_path)
        if not await self.conn.request(req_url):
            _LOGGER.warning("Error performing query.")
            return False
        _LOGGER.debug("ISY Query requested successfully.")
        return True

    async def send_x10_cmd(self, address, cmd):
        """
        Send an X10 command.

        address: String of X10 device address (Ex: A10)
        cmd: String of command to execute. Any key of x10_commands can be used
        """
        if cmd in X10_COMMANDS:
            command = X10_COMMANDS.get(cmd)
            req_url = self.conn.compile_url([CMD_X10, address, str(command)])
            result = await self.conn.request(req_url)
            if result is not None:
                _LOGGER.info("ISY Sent X10 Command: %s To: %s", cmd, address)
            else:
                _LOGGER.error("ISY Failed to send X10 Command: %s To: %s", cmd, address)
